import mock
import sys
import unittest

sys.path.append('..')
from wavefront.system_checker import SystemCheckerCommand

def literal(**kw):
    return namedtuple('literal', kw)(**kw)

class TestSystemCheckerCommand(unittest.TestCase):

    def test_execute_without_config(self):
        """
        Tests that execute() returns with error when --config is not provided
        """

        cmd = SystemCheckerCommand()
        args = {}
        with self.assertRaises(ValueError):
            cmd.execute(args)

    def test_execute_with_no_sections(self):
        cmd = SystemCheckerCommand()
        args = literal(config_file_path = '/tmp/system_checker.conf')
        m = mock.mock_open(read_data='[global]\n')
        with mock.patch('__main__.open', m, create=True):
            cmd.execute(args)
        
if __name__ == '__main__':
    unittest.main()



"""
This module calls the AWS ListMetrics() API followed by multiple calls to
GetMetricStatistics() to get metrics from AWS.

A dictionary configured by the 'metrics' key in the configuration file is
used to determine which metrics should lead to a call to GetMetricStatistics().

Each metric value returned from GetMetricStatistics() is sent to the Wavefront
proxy on port 2878 (or other port if configured differently).  Point tags
are picked up from the Dimensions.  Source is determined by searching
the point tags for a list of "accepted" source locations
(e.g., 'Service', 'LoadBalancerName', etc).

The last run time is stored in a configuration file in
/opt/wavefront/etc/aws-metrics.conf and will be used on the next run to
determine the appropriate start time.  If no configuration file is found,
the start time is determined by subtracting the delay_minutes from the
current time.
"""

import ConfigParser
import datetime
import json
import os
import os.path
import re

import logging.config

import dateutil

from wavefront.aws_common import AwsBaseMetricsCommand, AwsBaseMetricsConfiguration
from wavefront import utils

# Configuration for metrics that should be retrieved is contained in this
# configuration in a "metrics" key.  This is a dictionary
# where the key is a regular expression and the value is an object with keys:
#    * stats
#        a list of statistics to pull down with the GetMetricStatistics() call.
#        valid values are any of : 'Average', 'Maximum', 'Minimum', "SampleCount', 'Sum'
#    * source_names
#        an array of :
#          - tag names (Dimensions)
#          - Dimensions array index (0 based)
#          - String literals
#        The first match is returned as the source name.
#
# The key to the dictionary is a regular expression that should match a:
#     <namespace>.<metric_name> (lower case with /=>.)
#
DEFAULT_METRIC_CONFIG_FILE = './aws-metrics.json.conf'

# Mapping for statistic name to its "short" name.  The short name is used
# in the metric name sent to Wavefront
STAT_SHORT_NAMES = {
    'Average': 'avg',
    'Minimum': 'min',
    'Maximum': 'max',
    'Sum': 'sum',
    'SampleCount': 'count'
}

# characters to replace in the operation when creating the metric name
SPECIAL_CHARS_REPLACE_MAP = {
    '/': '-',
    ':': '-'
}

#pylint: disable=too-many-instance-attributes
class AwsCloudwatchConfiguration(object):
    """
    Configuration for Cloudwatch
    """

    def __init__(self, config, region):
        super(AwsCloudwatchConfiguration, self).__init__()

        self.config = config
        self.section_name = 'cloudwatch_' + region
        default_section_name = 'cloudwatch'

        self.enabled = self.config.getboolean(
            self.section_name, 'enabled', False, default_section_name)
        self.workers = int(self.config.get(
            self.section_name, 'workers', 1, default_section_name))
        self.has_suffix_for_single_stat = self.config.getboolean(
            self.section_name, 'single_stat_has_suffix', True,
            default_section_name)
        self.default_delay_minutes = int(self.config.get(
            self.section_name, 'first_run_start_minutes', 5,
            default_section_name))
        self.namespace = self.config.get(
            self.section_name, 'namespace', 'aws', default_section_name)
        self.ec2_tag_keys = self.config.getlist(
            self.section_name, 'ec2_tag_keys', [], default_section_name)
        self.metric_config_path = self.config.get(
            self.section_name, 'metric_config_path', DEFAULT_METRIC_CONFIG_FILE,
            default_section_name)

        self.start_time = self.config.getdate(
            self.section_name, 'start_time', None, default_section_name)
        self.end_time = self.config.getdate(
            self.section_name, 'end_time', None, default_section_name)
        self.update_start_end_times()

        self.namespaces = set()
        self.metrics_config = None

    def update_start_end_times(self):
        """
        Updates start/end times after last_run_time set
        """

        utcnow = (datetime.datetime.utcnow()
                  .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
        delta = datetime.timedelta(minutes=self.default_delay_minutes)
        last_run_time = self.config.get_last_run_time(self.section_name)
        if last_run_time:
            if not self.start_time or last_run_time > self.start_time:
                self.start_time = last_run_time - delta
                self.end_time = utcnow
        elif not self.start_time:
            self.start_time = utcnow - delta
            self.end_time = utcnow

    def set_last_run_time(self, run_time):
        """
        Sets the last run time to the run_time argument.

        Arguments:
        run_time - the time when this script last executed successfully (end)
        """

        self.config.set_last_run_time(run_time, self.section_name)

    def validate(self):
        """
        Validates configuration
        """
        if not self.metric_config_path:
            raise ValueError('options.metric_config_path is required')
        if not os.path.exists(self.metric_config_path):
            raise ValueError('ERROR: Configuration file (%s) does not exist' %
                             (self.metric_config_path))

    def load_metric_config(self):
        """
        Loads the metric configuration from the configuration file.
        """

        if self.metrics_config:
            return

        with open(self.metric_config_path, 'r') as conffd:
            config = json.load(conffd)

        if 'metrics' not in config:
            raise ValueError('ERROR: Configuration file (%s) is not valid' %
                             (self.metric_config_path))

        self.metrics_config = config['metrics']
        for _, config in self.metrics_config.iteritems():
            if 'namespace' in config and config['namespace']:
                self.namespaces.add(config['namespace'])

    #pylint: disable=unsupported-membership-test
    #pylint: disable=unsubscriptable-object
    def get_metric_config(self, namespace, metric_name):
        """
        Given a namespace and metric, get the configuration.

        Arguments:
        namespace - the namespace
        metric_name - the metric's name

        Returns:
        the configuration for this namespace and metric
        """

        self.load_metric_config()
        current_match = None
        metric = namespace.replace('/', '.').lower() + '.' + metric_name.lower()
        for name, config in self.metrics_config.iteritems():
            if re.match(name, metric, re.IGNORECASE):
                if current_match is None or \
                   ('priority' in current_match and \
                    current_match['priority'] < config['priority']):
                    current_match = config

        return current_match

#pylint: disable=too-many-instance-attributes
class AwsMetricsConfiguration(AwsBaseMetricsConfiguration):
    """
    Configuration file for this command
    """

    def __init__(self, config_file_path):
        super(AwsMetricsConfiguration, self).__init__(
            config_file_path=config_file_path)

        self.cloudwatch = {}
        for region in self.regions:
            self.cloudwatch[region] = AwsCloudwatchConfiguration(self, region)

    def get_region_config(self, region):
        """
        Gets the configuration for cloudwatch for the given region
        Arguments:
        region - the name of the region
        """

        if region in self.cloudwatch:
            return self.cloudwatch[region]
        else:
            return None

    def validate(self):
        """
        Checks that all required configuration items are set
        Throws:
        ValueError when a configuration item is missing a value
        """

        if (not self.aws_access_key_id or
                not self.aws_secret_access_key or
                not self.regions):
            raise ValueError('AWS access key ID, secret access key, '
                             'and regions are required')

        for _, cloudwatch in self.cloudwatch.iteritems():
            cloudwatch.validate()

class AwsCloudwatchMetricsCommand(AwsBaseMetricsCommand):
    """
    Wavefront command for retrieving metrics from AWS cloudwatch.
    """

    def __init__(self, **kwargs):
        super(AwsCloudwatchMetricsCommand, self).__init__(**kwargs)
        self.metrics_config = None

    def _initialize(self, arg):
        """
        Parses the arguments passed into this command.

        Arguments:
        arg - the argparse parser object returned from argparser
        """

        self.config = AwsMetricsConfiguration(arg.config_file_path)
        self.config.validate()
        try:
            logging.config.fileConfig(arg.config_file_path)
        except ConfigParser.NoSectionError:
            pass
        self.logger = logging.getLogger()

    #pylint: disable=no-self-use
    def get_help_text(self):
        """
        Returns help text for --help of this wavefront command
        """
        return "Pull metrics from AWS CloudWatch and push them into Wavefront"


    #pylint: disable=too-many-locals
    #pylint: disable=too-many-branches
    #pylint: disable=too-many-statements
    def _process_list_metrics_response(self, metrics, sub_account, region):
        """
        This function is called by _process_cloudwatch_region() after calling
        list_metrics() API.

        Loops over all metrics and call GetMetricStatistics() on each that are
        included by the configuration.

        Arguments:
        metrics - the array of metrics returned from ListMetrics() ('Metrics')
        sub_account - the AwsSubAccount object representing the top level
        """

        cloudwatch_config = self.config.get_region_config(region)
        start = cloudwatch_config.start_time
        end = cloudwatch_config.end_time
        session = sub_account.get_session(region)
        cloudwatch = session.client('cloudwatch')
        account_id = sub_account.get_account_id()

        for metric in metrics:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            top = (metric['Namespace']
                   .lower()
                   .replace('aws/', cloudwatch_config.namespace + '/')
                   .replace('/', '.'))
            metric_name = '{}.{}'.format(top, metric['MetricName'].lower())
            point_tags = {'Namespace': metric['Namespace'],
                          'Region': session.region_name,
                          'accountId': account_id}
            config = cloudwatch_config.get_metric_config(
                metric['Namespace'], metric['MetricName'])
            if config is None or len(config['stats']) == 0:
                self.logger.warning('No configuration found for %s/%s',
                                    metric['Namespace'], metric['MetricName'])
                continue

            dimensions = metric['Dimensions']
            for dim in dimensions:
                if (('dimensions_as_tags' in config and
                     dim['Name'] in config['dimensions_as_tags']) or
                        'dimensions_as_tags' not in config):
                    point_tags[dim['Name']] = dim['Value']
                if sub_account.instances and dim['Name'] == 'InstanceId':
                    instance_id = dim['Value']
                    region_instances = sub_account.get_instances(region)
                    if instance_id in region_instances:
                        instance_tags = region_instances[instance_id]
                        for key, value in instance_tags.iteritems():
                            point_tags[key] = value
                    else:
                        self.logger.warning('%s not found in region %s',
                                            instance_id, region)

            source, _ = AwsBaseMetricsCommand.get_source(
                config['source_names'], point_tags, dimensions)
            if not source:
                self.logger.warning('Source is not found in %s', str(metric))
                continue

            curr_start = start
            if (end - curr_start).total_seconds() > 86400:
                curr_end = curr_start + datetime.timedelta(days=1)
            else:
                curr_end = end

            while (curr_end - curr_start).total_seconds() > 0:
                if utils.CANCEL_WORKERS_EVENT.is_set():
                    break
                stats = cloudwatch.get_metric_statistics(
                    Namespace=metric['Namespace'],
                    MetricName=metric['MetricName'],
                    Dimensions=dimensions,
                    StartTime=curr_start,
                    EndTime=curr_end,
                    Period=60,
                    Statistics=config['stats'])

                number_of_stats = len(config['stats'])
                for stat in stats['Datapoints']:
                    for statname in config['stats']:
                        if utils.CANCEL_WORKERS_EVENT.is_set():
                            return
                        short_name = STAT_SHORT_NAMES[statname]
                        if (number_of_stats == 1 and
                                cloudwatch_config.has_suffix_for_single_stat):
                            full_metric_name = metric_name
                        else:
                            full_metric_name = metric_name + '.' + short_name

                        # remove point tags that we don't need for WF
                        if 'Namespace' in point_tags:
                            del point_tags['Namespace']

                        # send the metric to the proxy
                        tstamp = int(utils.unix_time_seconds(stat['Timestamp']))
                        self.proxy.transmit_metric(full_metric_name,
                                                   stat[statname],
                                                   tstamp,
                                                   source,
                                                   point_tags)

                curr_start = curr_end
                if (end - curr_start).total_seconds() > 86400:
                    curr_end = curr_start + datetime.timedelta(days=1)
                else:
                    curr_end = end

    def _process(self):

        # process each subaccount/region in parallel
        region_call_details = []
        for sub_account in self.account.get_sub_accounts():
            for region in self.account.regions:
                region_call_details.append((self._process_cloudwatch_region,
                                            (sub_account, region, )))

        self.logger.info('Processing %d region%s using %d threads',
                         len(self.account.regions),
                         's' if len(self.account.regions) > 1 else '',
                         len(self.account.regions))
        utils.parallel_process_and_wait(region_call_details,
                                        len(self.account.regions),
                                        self.logger)

    def _process_cloudwatch_region(self, sub_account, region):
        """
        Initialize and process a single region for a particular sub account.
        Response is paginated and each page is processed by its own thread
        Arguments:
        sub_account - the sub account
        region - the region name (us-west-1, etc)
        """

        cloudwatch_config = self.config.get_region_config(region)
        cloudwatch_config.update_start_end_times()
        cloudwatch_config.load_metric_config()
        self.logger.info('Loading metrics %s - %s (Region: %s, Namespace: %s)',
                         str(cloudwatch_config.start_time),
                         str(cloudwatch_config.end_time),
                         region,
                         ', '.join(cloudwatch_config.namespaces))

        function_pointers = []
        session = sub_account.get_session(region)
        cloudwatch = session.client('cloudwatch')
        for namespace in cloudwatch_config.namespaces:
            paginator = cloudwatch.get_paginator('list_metrics')
            if namespace == 'AWS/EC2':
                # for ec2 only: query with a filter for each instance
                # if you call list_metrics() on its own it returns several
                # instances that are no longer running
                instances = sub_account.get_instances(region)
                for instance_id in instances.instances:
                    dimensions = [{
                        'Name': 'InstanceId',
                        'Value': instance_id
                    }]
                    response = paginator.paginate(Namespace=namespace,
                                                  Dimensions=dimensions)
                    for page in response:
                        if utils.CANCEL_WORKERS_EVENT.is_set():
                            break
                        function_pointers.append(
                            (self._process_list_metrics_response,
                             (page['Metrics'], sub_account, region)))

            else:
                response = paginator.paginate(Namespace=namespace)
                for page in response:
                    if utils.CANCEL_WORKERS_EVENT.is_set():
                        break
                    function_pointers.append(
                        (self._process_list_metrics_response,
                         (page['Metrics'], sub_account, region)))

        if utils.CANCEL_WORKERS_EVENT.is_set():
            return
        self.logger.info('Metrics retrieved for region %s.  '
                         'Processing %d items in %d threads ...',
                         region, len(function_pointers),
                         cloudwatch_config.workers)
        utils.parallel_process_and_wait(function_pointers,
                                        cloudwatch_config.workers,
                                        self.logger)
        if not utils.CANCEL_WORKERS_EVENT.is_set():
            cloudwatch_config.set_last_run_time(cloudwatch_config.end_time)
            run_time = cloudwatch_config.config.get_last_run_time(
                cloudwatch_config.section_name)
            self.logger.info('Last run time updated to %s for %s',
                             str(run_time), region)



"""
This is a library for common code shared between the AWS Cloudwatch script
and the AWS Billing script commands.
"""

import datetime
import json
import os
import os.path

import boto3

from wavefront.metrics_writer import WavefrontMetricsWriter
from wavefront import command

# default configuration


# The directory where we should look for and store the cache
# files of instances and their tags.
CACHE_DIR = '/tmp'

#pylint: disable=too-many-instance-attributes
class AwsBaseMetricsConfiguration(command.CommandConfiguration):
    """
    Common Configuration file for this command
    """

    def __init__(self, config_file_path):
        super(AwsBaseMetricsConfiguration, self).__init__(
            config_file_path=config_file_path)

        self.writer_host = self.get('writer', 'host', '127.0.0.1')
        self.writer_port = int(self.get('writer', 'port', '2878'))
        self.is_dry_run = self.getboolean('writer', 'dry_run', True)

        self.aws_access_key_id = self.get('aws', 'access_key_id', None)
        self.aws_secret_access_key = self.get('aws', 'secret_access_key', None)
        self.regions = self.getlist('aws', 'regions', None, None, ',', True)
        self.sub_accounts = self.getlist('aws', 'sub_accounts', [])

        self._setup_output(self)

class AwsBaseMetricsCommand(command.Command):
    """
    Abstract base class for both AWS cloudwatch metrics and AWS billing metrics
    commands.
    """

    def __init__(self, **kwargs):
        super(AwsBaseMetricsCommand, self).__init__(**kwargs)

        self.account = None
        self.config = None
        self.proxy = None

    def _init_proxy(self):
        """
        Initializes the proxy writer
        """

        self.proxy = WavefrontMetricsWriter(self.config.writer_host,
                                            self.config.writer_port,
                                            self.config.is_dry_run)
        self.proxy.start()

    def _process(self):
        """
        Process this aws command (implemented by derived class)
        """

        pass

    def _execute(self):
        """
        Execute this command
        """

        self._init_proxy()
        self.account = AwsAccount(self.config, True)
        self._process()

    @staticmethod
    def get_source(source_names, point_tags, dimensions=None):
        """
        Determine the source from the point tags.
        Argument:
        source_names - the key names in priority order to use as source
        point_tags - all the point tags for this metric (dictionary)
        dimensions - the dimensions for this metric (list of objects)

        Returns:
        Tuple of (source value, key of the source of the source)
        """

        for name in source_names:
            if name[0:1] == '=':
                return (name[1:], None)

            if name in point_tags and point_tags[name]:
                return (point_tags[name], name)

            if dimensions:
                for dim in dimensions:
                    if dim['Name'] == name and dim['Value']:
                        return (dim['Value'], name)

        return (None, None)

class AwsAccount(object):
    """
    Represents the AWS account and all of its sub accounts
    """

    def __init__(self, config, load=False):
        super(AwsAccount, self).__init__()
        self.config = config
        self.regions = self.config.regions
        self.sub_accounts = []
        self.sessions = {}
        if load:
            for sub_account in self.get_sub_accounts():
                sub_account.load_ec2_instance_data()

    def get_sub_accounts(self):
        """
        Gets a list of sub accounts
        """
        if not self.sub_accounts:
            for sub_account in self.config.sub_accounts:
                self.sub_accounts.append(AwsSubAccount(self, sub_account))
        return self.sub_accounts

    def get_account_id(self, role_arn=None):
        """
        Gets the account id by either parsing it from the role ARN or by
        getting the currently logged in user's ARN and parsing from there.
        """

        if role_arn:
            arn = role_arn

        else:
            iam_client = self.get_session(
                self.regions[0], None, None).client('iam')
            arn = iam_client.get_user()['User']['Arn']

        return arn.split(':')[4]

    def get_session(self, region, role_arn, external_id):
        """
        Creates a new session object in the given region
        Arguments:
        region - the region name
        """

        if role_arn:
            if not external_id:
                raise ValueError('External ID is required for region ' +
                                 region + ' and role ARN ' + role_arn)

        access_key_id = self.config.aws_access_key_id
        secret_access_key = self.config.aws_secret_access_key
        if role_arn:
            session = boto3.session.Session()
            client = session.client(
                'sts',
                region_name=region,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key)
            role = client.assume_role(RoleArn=role_arn,
                                      ExternalId=external_id,
                                      RoleSessionName='wavefront_session')
            return boto3.Session(
                role['Credentials']['AccessKeyId'],
                role['Credentials']['SecretAccessKey'],
                role['Credentials']['SessionToken'],
                region_name=region)

        else:
            return boto3.session.Session(
                region_name=region,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key)


#pylint: disable=too-few-public-methods
class AwsSubAccountConfiguration(object):
    """
    Configuration for a specific sub account section in the INI file
    """

    def __init__(self, config, section_name):
        super(AwsSubAccountConfiguration, self).__init__()

        self.config = config
        self.enabled = self.config.getboolean(section_name, 'enabled', False)
        self.role_arn = self.config.get(section_name, 'role_arn', None)
        self.role_external_id = self.config.get(section_name, 'external_id', None)

class AwsSubAccount(object):
    """
    AWS sub-account
    """

    def __init__(self, parent, name):
        super(AwsSubAccount, self).__init__()

        self.parent_account = parent
        section_name = 'aws_sub_account_' + name
        self.sub_account_config = AwsSubAccountConfiguration(
            parent.config, section_name)

        self.instances = {}

    def get_account_id(self):
        """
        Gets the account id by either parsing it from the role ARN or by
        getting the currently logged in user's ARN and parsing from there.
        """

        return self.parent_account.get_account_id(
            self.sub_account_config.role_arn)

    def load_ec2_instance_data(self):
        """
        Loads all AWS EC2 instances and related tags in the account's regions
        Arguments:
        """

        for region in self.parent_account.regions:
            reg_config = (self.parent_account.config.get_region_config(region))
            ec2_tag_keys = reg_config.ec2_tag_keys
            self.instances[region] = AwsInstances(
                self, region, ec2_tag_keys, True)

    def get_instances(self, region):
        """
        Gets the instances for the given region
        Arguments:
        region - the region name
        Returns:
        AwsInstances object for the given region or None
        """

        if region in self.instances:
            return self.instances[region]
        else:
            return None

    def get_session(self, region):
        """
        Creates a new session object in the given region
        Arguments:
        region - the region name
        check_cache - True to check the cache before creating new session
        """

        return self.parent_account.get_session(
            region, self.sub_account_config.role_arn,
            self.sub_account_config.role_external_id)

#pylint: disable=too-few-public-methods
class AwsInstances(object):
    """
    Queries and caches the tags of all instances in a region.  Results are
    cached in a configured directory.  Cached results are used if the
    date of the file is within the last day (using modified time).
    The configuration object stores the AWS tag keys to retrieve from each
    instance.  If this configuration is not set (blank or null), this
    class does nothing.
    """
    def __init__(self, sub_account, region, ec2_tag_keys, load_now=False):
        """
        Initializes the class.
        Arguments:
        sub_account -
        region - the region name
        ec2_tag_keys - array of tag key names
        load_now -
        """

        super(AwsInstances, self).__init__()
        self.sub_account = sub_account
        self.region = region
        self.ec2_tag_keys = ec2_tag_keys
        self.instances = None
        if load_now:
            self.load()

    def _get_cache_file_path(self):
        """
        Generates a file path for the given account
        Arguments:
        sub_account - the account
        """
        fname = ('instance_tag_%s_cache_%s.json' %
                 (self.sub_account.get_account_id(), self.region, ))
        return os.path.join(CACHE_DIR, fname)

    def _query_instance_tags(self):
        """
        Calls EC2.DescribeInstances() and retrieves all instances and their tags
        """

        self.instances = {}

        _instances = (self.sub_account.get_session(self.region)
                      .resource('ec2').instances.all())
        for instance in _instances:
            tags = {}

            # hard-coded instance attributes (data coming from instance object)
            if 'instanceType' in self.ec2_tag_keys:
                tags['instanceType'] = instance.instance_type
            if 'imageId' in self.ec2_tag_keys:
                tags['imageId'] = instance.instance_type
            if 'publicDnsName' in self.ec2_tag_keys:
                tags['publicDnsName'] = instance.public_dns_name
            if 'privateDnsName' in self.ec2_tag_keys:
                tags['privateDnsName'] = instance.private_dns_name
            if 'vpcId' in self.ec2_tag_keys:
                tags['vpcId'] = instance.vpc_id
            if 'architecture' in self.ec2_tag_keys:
                tags['architecture'] = instance.architecture

            # tags coming from the EC2 tags
            if instance.tags:
                for tag in instance.tags:
                    if (self.ec2_tag_keys[0] == '*' or
                            tag['Key'] in self.ec2_tag_keys):
                        tags[tag['Key']] = tag['Value']

            # store the tags in the dictionary
            self.instances[instance.id] = tags

        # store the results on disk for next time
        with open(self._get_cache_file_path(), 'w') as cachefd:
            json.dump(self.instances, cachefd)

    def _load_instance_tags_from_cache(self):
        """
        Loads the tags from the cache file if it exists.
        Returns:
        True - when data loaded from cache; False - o/w
        """

        path = self._get_cache_file_path()
        if os.path.exists(path):
            now = datetime.datetime.utcnow()
            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
            time_to_refresh = mtime + datetime.timedelta(days=1)
            if now > time_to_refresh:
                with open(path, 'r') as contents:
                    self.instances = json.load(contents)
                    return True

        return False

    def load(self):
        """
        Loads the instances and their tags.  Caches that data for at most one
        day (configurable?).
        """

        if self.instances or not self.ec2_tag_keys:
            return
        if not self._load_instance_tags_from_cache():
            self._query_instance_tags()

    def __contains__(self, item):
        if not self.instances:
            return False
        return item in self.instances
    def __getitem__(self, item):
        if not self.instances:
            return None
        return self.instances[item]



#!/usr/bin/env python
"""
This script pulls metrics from New Relic.  See the README for
more details on configuration.
"""

import ConfigParser
import datetime
import hashlib
import json
import numbers
import os
import os.path
import re
import time

import logging.config
import dateutil.parser

from wavefront.utils import parallel_process_and_wait
from wavefront import command, utils
from wavefront.newrelic_common import NewRelicCommand

# default number of metric names to include with each GetMetricData
# API call
DEFAULT_MAX_METRIC_NAME_COUNT = 25

# default location for the configuration file.
DEFAULT_CONFIG_FILE_PATH = '/opt/wavefront/etc/wavefront-collector-newrelic.conf'

#pylint: disable=too-many-instance-attributes
class NewRelicPluginConfiguration(command.CommandConfiguration):
    """
    Stores the configuration for this plugin
    """

    #pylint: disable=too-many-statements
    def __init__(self, config_file_path):
        super(NewRelicPluginConfiguration, self).__init__(
            config_file_path=config_file_path)

        self.api_key = self.get('api', 'key', '')
        self.api_endpoint = self.get(
            'api', 'endpoint', 'https://api.newrelic.com/v2')
        self.api_key_header_name = self.get(
            'api', 'key_header_name', 'X-Api-Key')
        self.api_log_path = self.get('api', 'log_path', None)

        self.fields = self.getlist('filter', 'names', [])
        self.fields_regex = self.getlist('filter', 'regex', [])
        self.fields_regex_compiled = []
        for regex in self.fields_regex:
            self.fields_regex_compiled.append(re.compile(regex))
        self.fields_blacklist_regex = self.getlist(
            'filter', 'blacklist_regex', [])
        self.fields_blacklist_regex_compiled = []
        for regex in self.fields_blacklist_regex:
            self.fields_blacklist_regex_compiled.append(re.compile(regex))
        self.additional_fields = self.getlist('filter', 'additional_fields', [])
        self.application_ids = self.getlist('filter', 'application_ids', [])
        self.start_time = self.getdate('filter', 'start_time', None)
        self.end_time = self.getdate('filter', 'end_time', None)
        last_run_time = self.get_last_run_time()
        if self.start_time and self.end_time and last_run_time:
            if last_run_time > self.start_time:
                self.start_time = last_run_time
        elif last_run_time:
            self.start_time = last_run_time

        self.include_application_summary = self.getboolean(
            'options', 'include_application_summary', True)
        self.include_host_app_summary = self.getboolean(
            'options', 'include_host_application_summary', True)
        self.include_server_summary = self.getboolean(
            'options', 'include_server_summary', True)
        self.include_servers = self.getboolean(
            'options', 'include_server_details', False)
        self._setup_output(self)

        self.include_hosts = self.getboolean('options', 'include_hosts', True)
        self.min_delay = int(self.get('options', 'min_delay', 60))
        self.wf_api_key = self.get('wavefront_api', 'key', '')
        self.wf_api_endpoint = self.get(
            'wavefront_api', 'endpoint', 'https://metrics.wavefront.com/')
        self.skip_null_values = self.getboolean(
            'options', 'skip_null_values', False)
        self.default_null_value = self.get('options', 'default_null_value', 0)
        self.max_metric_names = int(self.get(
            'options', 'max_metric_names', DEFAULT_MAX_METRIC_NAME_COUNT))
        self.delay = int(self.get('options', 'delay', 0))
        self.writer_host = self.get('writer', 'host', '127.0.0.1')
        self.writer_port = int(self.get('writer', 'port', '2878'))
        self.is_dry_run = self.getboolean('writer', 'dry_run', True)

        self.query_api_key = self.get('query_api', 'key', '')
        self.query_api_endpoint = self.get(
            'query_api', 'endpoint', 'https://insights-api.newrelic.com/v1')
        self.query_api_key_header_name = self.get(
            'query_api', 'key_header_name', 'X-Query-Key')

        self.queries = {}
        for section in self.config.sections():
            if section[0:5].lower() == 'query':
                if self.getboolean(section, 'enabled', True):
                    name = self.get(section, 'name', 'query')
                    query = self.get(section, 'query', '')
                    self.queries[name] = query
        self.workers = int(self.get('options', 'workers', 30))
        self.cache_directory = '/tmp/wfnrcache'
        if not os.path.exists(self.cache_directory):
            os.makedirs(self.cache_directory)

        self.metric_last_sent = {}
        self.send_zero_every = int(self.get('options', 'send_zero_every', 0))

    def get_value_to_send(self, name, value):
        """
        Gets the value to send for the given value.  This allows us to translate
        the value from "NaN" to "0", for example.
        This function returns None when it should not be sent.
        Arguments:
        name - the name of the metric
        value - the value to check

        Return:
        The value to send or None to not send
        """

        if (not value or
                str(value).lower() == 'nan' or
                not isinstance(value, numbers.Number)):
            value = self.default_null_value

        if self.send_zero_every:
            current = (datetime.datetime.utcnow()
                       .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
            if value == 0 and name in self.metric_last_sent:
                last_sent = self.metric_last_sent[name]
                if (current - last_sent).total_seconds() < self.send_zero_every:
                    value = None
            self.metric_last_sent[name] = current

        return value

    def validate(self):
        """
        Checks that all required configuration items are set
        Throws:
        ValueError when a configuration item is missing a value
        """

        if not self.api_key:
            raise ValueError('api.key configuration is required')
        if not self.api_endpoint:
            raise ValueError('api.endpoint configuration is required')

class NewRelicMetricRetrieverCommand(NewRelicCommand):
    """
    Command object for retrieving New Relic metrics via the REST API.
    """

    def __init__(self, **kwargs):
        super(NewRelicMetricRetrieverCommand, self).__init__(**kwargs)
        self.description = 'New Relic Metric Retriever'

        # cache of metric names for given host-app
        # Key is application_id-host_id
        # Value is a dictionary:
        #   last_refresh - a datetime object in UTC of the last time value
        #                  is refreshed
        #   value - the list of metric names (list of strings)
        self.metric_name_cache = {}

    #pylint: disable=no-self-use
    def get_help_text(self):
        """
        Help text for this command.
        """

        return "Pull metrics from New Relic"

    def _initialize(self, args):
        """
        Parses the arguments passed into this command.

        Arguments:
        arg - the argparse parser object returned from argparser
        """

        self.config = NewRelicPluginConfiguration(args.config_file_path)
        self.config.validate()
        try:
            logging.config.fileConfig(args.config_file_path)
        except ConfigParser.NoSectionError:
            pass
        self.logger = logging.getLogger()

    def get_metric_names_for_path(self, path, names_filter):
        """
        Gets the metric names for the given URL path.

        Arguments:
        path - the URL path
        names_filter - list of names to filter (additional calls)

        Returns:
        A list of string metric names
        """

        self.logger.debug('Getting metric names for %s', path)

        if utils.CANCEL_WORKERS_EVENT.is_set():
            return []

        # get a hash value instead of using path so we can store on disk
        hashval = hashlib.md5(path).hexdigest()
        filepath = self.config.cache_directory + '/' + hashval

        # cache this list so we only call once per day
        now = datetime.datetime.utcnow()
        if hashval not in self.metric_name_cache:
            if os.path.exists(filepath):
                with open(filepath, 'r') as contents:
                    names = json.load(contents)
                    self.metric_name_cache[hashval] = {
                        'value': names,
                        'last_refresh': datetime.datetime.fromtimestamp(
                            os.path.getmtime(filepath))
                    }

        if hashval in self.metric_name_cache:
            last_refresh = self.metric_name_cache[hashval]['last_refresh']
            time_to_refresh = last_refresh + datetime.timedelta(days=1)
            self.logger.debug('Checking %s; Last Refresh: %s; '
                              'Time to Refresh: %s',
                              path, str(last_refresh), str(time_to_refresh))
            if now < time_to_refresh:
                return self.metric_name_cache[hashval]['value']

            else:
                self.logger.debug('Refreshing metric names for %s; '
                                  'Last Refresh: %s; Time to Refresh: %s',
                                  path, str(last_refresh),
                                  str(time_to_refresh))

        names = []
        response = self.call_paginated_api(
            path + '/metrics.json', None, None, None)
        for page in response:
            metrics = page['metrics']
            for metric in metrics:
                names.append(metric['name'])

        for name in names_filter:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                return []
            response = self.call_paginated_api(
                path + '/metrics.json', {"name": name}, None, None)
            for page in response:
                metrics = page['metrics']
                for metric in metrics:
                    names.append(metric['name'])

        self.metric_name_cache[hashval] = {
            'last_refresh': now,
            'value': names
        }
        with open(filepath, 'w') as contents:
            json.dump(names, contents)

        return names

    def _server_metrics(self, start, end):
        """
        Pull the server metrics from New Relic and post them to the WF proxy.

        Arguments:
        start - the start datetime object
        end - the end datetime object
        """

        if (not self.config.include_servers or
                utils.CANCEL_WORKERS_EVENT.is_set()):
            return

        self.logger.info('Retrieving server metrics ...')
        servers = self.call_api('/servers.json')[0]
        for server in servers['servers']:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                return

            server_id = server['id']
            server_name = server['name']
            tags = {
                'server_id': server_id,
                'server_name': server_name
            }
            if (self.config.include_server_summary and
                    'summary' in server):
                summary = server['summary']
                for key, value in summary.iteritems():
                    if utils.CANCEL_WORKERS_EVENT.is_set():
                        break
                    metric_name = ('servers/%s/%s' % (server_name, key))
                    self.send_metric(self.proxy, metric_name, value, 'newrelic',
                                     server['last_reported_at'], tags,
                                     self.config.get_value_to_send, self.logger)
            self.send_metrics_for_server(server_id, server_name, start, end)

    def _application_metrics(self, start, end):
        """
        Pull the app metrics from New Relic and post them to the WF proxy
        Arguments:
        start - the start datetime object
        end - the end datetime object
        """

        if (not self.config.include_application_summary and
                not self.config.include_host_app_summary and
                not self.config.include_hosts):
            return

        query_string = None
        if self.config.application_ids:
            query_string = {
                'filter[ids]': ','.join(self.config.application_ids)
            }

        args = (start, end)
        self.call_paginated_api('/applications.json', query_string,
                                self._handle_applications_response, args)

    #pylint: disable=too-many-branches
    def _handle_applications_response(self, response, start, end):
        """
        Function that handles each page response that is returned from the
        /applications.json request
        Arguments:
        response - the page's response
        start - the start datetime object
        end - the end datetime object
        """

        if not response or 'applications' not in response:
            return

        for app in response['applications']:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                return
            self.logger.info('Retrieving %s - %s (app: %s)',
                             str(start), str(end), app['name'])
            if not app['reporting']:
                continue

            app_id = app['id']
            app_name = app['name']
            tags = {
                'app_id': app_id,
                'app_name': app_name
            }
            if self.config.include_application_summary:
                self.logger.info('Retrieving application summary (%s)...',
                                 app['last_reported_at'])
                summary = app['application_summary']
                for key, value in summary.iteritems():
                    if utils.CANCEL_WORKERS_EVENT.is_set():
                        break
                    metric_name = 'apps/%s/%s' % (app_name, key)
                    self.send_metric(self.proxy, metric_name, value,
                                     'newrelic', app['last_reported_at'],
                                     tags, self.config.get_value_to_send,
                                     self.logger)

                if 'end_user_summary' in app:
                    summary = app['end_user_summary']
                    for key, value in summary.iteritems():
                        if utils.CANCEL_WORKERS_EVENT.is_set():
                            break
                        metric_name = 'apps/%s/enduser/%s' % (app_name, key)
                        self.send_metric(self.proxy, metric_name, value,
                                         'newrelic', app['last_reported_at'],
                                         tags, self.config.get_value_to_send,
                                         self.logger)

            if ((self.config.include_hosts or
                 self.config.include_host_app_summary) and
                    not utils.CANCEL_WORKERS_EVENT.is_set()):

                for host_id in app['links']['application_hosts']:
                    if utils.CANCEL_WORKERS_EVENT.is_set():
                        break
                    self.send_metrics_for_host(app_id, app_name,
                                               host_id, start, end)

    #pylint: disable=too-many-branches
    def _execute(self):
        """
        Makes API calls to New Relic on behalf of the configured user and
        reports metrics to configured endpoint.
        """

        try:
            self.init_proxy()
            # construct start time for when to get metrics starting from
            if self.config.start_time:
                start = self.config.start_time
            else:
                start = ((datetime.datetime.utcnow() -
                          datetime.timedelta(seconds=60.0))
                         .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
            if self.config.end_time:
                end = self.config.end_time
            else:
                end = None

            if start is not None:
                if end is None:
                    end = (datetime.datetime.utcnow()
                           .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))

                if (end - start).total_seconds() < self.config.min_delay:
                    self.logger.info('Not running since %s - %s < 60',
                                     str(end), str(start))
                    return

            start = start.replace(microsecond=0, tzinfo=dateutil.tz.tzutc())
            self.logger.info('Running %s - %s', str(start), str(end))

            # if the time is more than 10 minutes, NR will make the sample size
            # larger than a minute.  so, we'll grab the data in chunks
            # (10m at a time)
            curr_start = start
            if not end:
                end = (datetime.datetime.utcnow()
                       .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))
            curr_end = end
            curr_diff = curr_end - curr_start
            while (curr_diff.total_seconds() > 0 and
                   not utils.CANCEL_WORKERS_EVENT.is_set()):
                if (curr_diff.total_seconds() > 600 or
                        curr_diff.total_seconds() < 60):
                    curr_end = curr_start + datetime.timedelta(minutes=10)

                # get the application metrics
                self._application_metrics(curr_start, curr_end)

                # get the servers
                self._server_metrics(curr_start, curr_end)

                # save "last run time" and update curr_* variables
                self.config.set_last_run_time(curr_end)
                curr_start = curr_end
                curr_end = end
                curr_diff = curr_end - curr_start
                if (curr_diff.total_seconds() > 600 and
                        not utils.CANCEL_WORKERS_EVENT.is_set()):
                    time.sleep(30)

        finally:
            self.config.start_time = None
            self.config.end_time = None

    #pylint: disable=too-many-arguments
    #pylint: disable=too-many-locals
    #pylint: disable=too-many-branches
    def send_metrics_for_host(self, app_id, app_name, host_id, start, end):
        """
        Retrieves the metrics for the given host and sends them to configured
        endpoint.

        Arguments:
        app_id - The application's ID
        app_name - The application's name
        host_id - The host ID
        host_name - the host name
        start - the starting datetime object
        end - the ending datetime object
        """

        self.logger.info('Retrieving host %s summary metrics ...', str(host_id))

        app_host = self.call_api(
            '/applications/%s/hosts/%s.json' % (app_id, host_id))[0]
        if not app_host:
            return

        app_host = app_host['application_host']
        host_name = app_host['host']
        tags = {
            'app_id': app_id,
            'app_name': app_name
        }
        self.logger.debug('host: %s', host_name)
        if (self.config.include_host_app_summary and
                'application_summary' in app_host):
            summary = app_host['application_summary']
            for key, value in summary.iteritems():
                metric_name = 'apps/%s/%s' % (app_name, key)
                self.send_metric(self.proxy, metric_name, value, host_name,
                                 start.isoformat(), tags,
                                 self.config.get_value_to_send, self.logger)

        if not self.config.include_hosts:
            return

        self.logger.info('Retrieving host %s metrics ...', host_name)
        path = '/applications/%s/hosts/%s' % (app_id, host_id)

        # get the list of names from the config and by querying NR
        if self.config.fields:
            names = list(self.config.fields)
        else:
            names = []
        names.extend(self.get_metric_names_for_path(
            path, self.config.additional_fields))

        # now we have a list of metric names apply some filtering based
        # on the white and black list
        if self.config.fields_regex:
            for fieldname in names[:]:
                found = False
                for pattern in self.config.fields_regex_compiled:
                    if pattern.match(fieldname):
                        found = True
                        break
                if not found:
                    names.remove(fieldname)
        # black list ...
        if self.config.fields_blacklist_regex:
            self.logger.debug('Checking field names against black list')
            for fieldname in names[:]:
                found = False
                for pattern in self.config.fields_blacklist_regex_compiled:
                    if pattern.match(fieldname):
                        found = True
                        break
                if found:
                    names.remove(fieldname)

        self.logger.debug('%d Metric names for path %s:\n%s',
                          len(names), path, str(names))
        self.get_metrics_for_path(path, names, start, end, host_name, tags)

    def send_metrics_for_server(self, server_id, server_name, start, end):
        """
        Retrieves the metrics for the given server and sends them to configured
        endpoint.

        Arguments:
        server_id - The server's ID
        start - the starting datetime object
        end - the ending datetime object
        """

        path = '/servers/%s' % (server_id)
        fields = self.get_metric_names_for_path(path, [])
        tags = {
            'server_id': server_id,
            'server_name': server_name
        }
        self.get_metrics_for_path(path, fields, start, end, server_name, tags)

    #pylint: disable=too-many-arguments
    def get_metrics_for_path(self, path, fields_to_get, start, end,
                             src_name, tags):
        """
        Get metrics from the NR API and send them to Wavefront.

        Arguments:
        path - API path
        fields_to_get - the list of field names to retrieve
        start -
        end -
        src_name -
        tags
        """

        if utils.CANCEL_WORKERS_EVENT.is_set():
            return

        self.logger.debug('Getting metrics for path %s (src=%s, tags=%s)',
                          path, src_name, str(tags))

        # there is a limit on the size of the query string.  we pick a
        # max number of metric names to get on each request and loop
        # until we've got all metric values.
        # fields_to_get_temp contains all the remaining metric names
        # to get. as each group of metric names is retrieved, they are
        # removed (see end of loop).
        # during each iteration, get the first X (max) metric values.
        fields_to_get_temp = list(fields_to_get)

        function_pointers = []
        while len(fields_to_get_temp) > 0:
            query_string = {
                'from': start.isoformat(),
                'to': end.isoformat(),
                'names[]': fields_to_get_temp[0:self.config.max_metric_names],
                'raw': True,
                'summarize': False
            }
            del fields_to_get_temp[0:self.config.max_metric_names]

            function_pointers.append((self.response_worker,
                                      (path, query_string, src_name, tags)))

        self.logger.debug('Metrics for path %s: work queue has %d items',
                          path, len(function_pointers))
        parallel_process_and_wait(function_pointers, self.config.workers,
                                  self.logger)

    def response_worker(self, path, query_string, src_name, tags):
        """
        Worker that runs in a separate thread to execute the call_api call
        and process the response.
        """

        response = self.call_api(path + '/metrics/data.json', query_string)[0]
        if not response or 'metric_data' not in response:
            self.logger.warning('%s: response does not contain metric_data',
                                path)
            return

        # need to get a new writer for each thread (create a new socket)
        writer = NewRelicCommand.get_writer_from_config(self.config)

        try:
            # parse and collect the metrics
            metrics = response['metric_data']['metrics']
            for metric_detail in metrics:
                if utils.CANCEL_WORKERS_EVENT.is_set():
                    break
                name = metric_detail['name']
                for time_slice in metric_detail['timeslices']:
                    for metric, value in time_slice['values'].iteritems():
                        if utils.CANCEL_WORKERS_EVENT.is_set():
                            break
                        metric_name = '%s/%s' % (name, metric)
                        NewRelicCommand.send_metric(
                            writer, metric_name, value, src_name,
                            time_slice['to'], tags,
                            self.config.get_value_to_send)
        finally:
            writer.stop()



"""
Setup script for the Wavefront collector tools.
"""

import setuptools
import setuptools.command.install


# The easiest way to convert the markdown to RestructuredText is to use
# pandoc.  There is a Python frontend to that package called pypandoc.
# To use this code you will need to :
#   1. Download and install pandoc  (http://pandoc.org/installing.html)
#   2. pip install pypandoc
# see: https://coderwall.com/p/qawuyq/use-markdown-readme-s-in-python-modules
try:
    import pypandoc
    LONG_DESCRIPTION = pypandoc.convert_file(source_file='README.md',
                                             format='markdown_github',
                                             to='rst',
                                             extra_args=['-s', '--columns=1000'])
except (IOError, ImportError):
    LONG_DESCRIPTION = ''

setuptools.setup(
    name='wavefront_collector',
    version='0.0.44',
    author='Wavefront',
    author_email='mike@wavefront.com',
    description=('Wavefront Collector Tools'),
    license='BSD',
    long_description=LONG_DESCRIPTION,
    keywords='wavefront wavefront_integration collector metrics',
    url='https://www.wavefront.com',
    install_requires=['wavefront_client', 'python-dateutil', 'logging',
                      'python-daemon', 'boto3', 'ndg-httpsclient'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Utilities',
        'License :: OSI Approved :: BSD License',
    ],
    package_data={'wavefront': ['data/*']},
    packages=['wavefront'],
    scripts=['wf', 'wavefront-collector']
)



"""
This module handles parsing the AWS Billing Reports (stored on S3 in .zip
or just plain .csv format) and creating metrics to be sent to the WF proxy.
"""

import ConfigParser
import datetime
import io
import os
import sys
import time
import traceback
import zipfile

import logging.config

import dateutil

from wavefront.aws_common import AwsBaseMetricsCommand, AwsBaseMetricsConfiguration
from wavefront import utils

#pylint: disable=too-few-public-methods
#pylint: disable=too-many-instance-attributes
class AwsBillingConfiguration(AwsBaseMetricsConfiguration):
    """
    Configuration for billing
    """
    def __init__(self, config_file_path):
        super(AwsBillingConfiguration, self).__init__(
            config_file_path=config_file_path)

        self.enabled = self.getboolean('aws_billing', 'enabled', False)
        self.role_arn = self.get('aws_billing', 'role_arn', None)
        self.role_external_id = self.get(
            'aws_billing', 'external_id', None)
        self.billing_thread_names = self.getlist(
            'aws_billing', 'billing_threads', [])
        self.ec2_tag_keys = self.getlist('aws_billing', 'ec2_tag_keys', [])
        self.billing_threads = []
        for name in self.billing_thread_names:
            section = 'billing-' + name
            self.billing_threads.append(
                AwsBillingDetailThreadConfiguration(self, section))

    def validate(self):
        """
        Validation of configuration
        """

        pass

    def get_region_config(self, _):
        """
        Gets the configuration for cloudwatch for the given region
        Arguments:
        region - the name of the region
        """

        return self

#pylint: disable=too-few-public-methods
#pylint: disable=too-many-instance-attributes
class AwsBillingDetailThreadConfiguration(object):
    """
    Configuration for a billing detail section in the configuration file
    """

    def __init__(self, config, section_name):
        super(AwsBillingDetailThreadConfiguration, self).__init__()

        self.config = config
        self.section_name = section_name
        self.last_run_time_section = section_name
        self.tmp_dir = self.config.get(section_name, 'tmp_dir', '/tmp/')
        self.namespace = self.config.get(section_name, 'namespace', None)
        self.enabled = self.config.getboolean(section_name, 'enabled', False)
        self.region = self.config.get(section_name, 's3_region', None)
        self.bucket = self.config.get(section_name, 's3_bucket', None)
        self.prefix = self.config.get(section_name, 's3_prefix', None)
        self.header_row_index = int(
            self.config.get(section_name, 'header_row_index', 1))
        self.dimensions = self._build_table(
            self.config.getlist(section_name, 'dimension_column_names', []))
        self.metrics = self._build_table(
            self.config.getlist(section_name, 'metric_column_names', []))
        self.source_names = self.config.getlist(section_name, 'source_names', [])
        self.dates = self._build_table(
            self.config.getlist(section_name, 'date_column_names', []), '|')
        self.duration = self.config.getlist(section_name, 'duration_column_names', [])
        self.instance_id_columns = self.config.getlist(
            section_name, 'instance_id_column_names', [])
        self.delay = int(self.config.get(section_name, 'delay', 3600))
        self.record_id_column = self.config.get(
            section_name, 'record_id_column_name', None)
        self.maximum_number_of_rows = int(self.config.get(
            section_name, 'maximum_number_of_rows', 0))
        self.sleep_after_rows = int(self.config.get(
            section_name, 'sleep_after_rows', 0))
        self.sleep_ms = float(self.config.get(
            section_name, 'sleep_ms', 0.0)) / 1000

    @staticmethod
    def _build_table(lst, delimiter=':'):
        """
        Build a dictionary from a list of delimiter-separated key-value pairs
        Arguments:
        lst - list of strings
        delimiter - delimiter between components of each string in the lst

        Returns:
        dictionary with the key being the string on the left side of
        the delimiter and the value of the dictionary key being the string
        on the right side
        """

        rtn = {}
        if lst:
            for item in lst:
                parts = item.split(delimiter)
                if len(parts) == 1:
                    rtn[parts[0]] = parts[0]
                elif len(parts) == 2:
                    rtn[parts[0]] = parts[1]

        return rtn

    def get_last_record_id(self, curr_month):
        """
        Gets the last record id for the given month
        """
        return self.config.output.get(
            self.section_name, 'last_record_id_' + curr_month, None)

    def set_last_record_id(self, curr_month, record_id):
        """
        Sets the last record id read

        Arguments:
        record_id - last record id
        """

        if not record_id:
            return

        self.config.output.set(
            self.section_name, 'last_record_id_' + curr_month, record_id)
        self.config.output.save()

class AwsBillingMetricsCommand(AwsBaseMetricsCommand):
    """
    Billing metrics command object.  Grabs metrics from billing CSV files.
    """

    def __init__(self, **kwargs):
        super(AwsBillingMetricsCommand, self).__init__(**kwargs)

    def _initialize(self, args):
        """
        Initialize this command

        Arguments:
        arg - the argparse parser object returned from argparser
        """

        self.config = AwsBillingConfiguration(args.config_file_path)
        self.config.validate()
        try:
            logging.config.fileConfig(args.config_file_path)
        except ConfigParser.NoSectionError:
            pass
        self.logger = logging.getLogger()

    def _process(self):
        """
        Processes the latest billing details CSV file.  A few helpful sites:
        http://www.dowdandassociates.com/products/cloud-billing/documentation/1.0/schema/
        http://docs.aws.amazon.com/awsaccountbilling/latest/aboutv2/detailed-billing-reports.html#reportstagsresources
        """

        utcnow = (datetime.datetime.utcnow()
                  .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))

        if utils.CANCEL_WORKERS_EVENT.is_set():
            return

        if not self.config.enabled:
            self.logger.info('Billing is disabled')
            return

        for config in self.config.billing_threads:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            try:
                if config.enabled:
                    last_run_time = config.config.get_last_run_time()
                    if last_run_time:
                        diff = utcnow - last_run_time
                        if diff.total_seconds() <= config.delay:
                            self.logger.info('Not ready to run %s (last run at '
                                             '%s; expected delay interval is %ds)',
                                             config.section_name,
                                             str(last_run_time),
                                             config.delay)
                            continue
                    if config.bucket == 'local':
                        self.logger.info('Running in local mode ...')
                        self._get_csv_from_local(config)

                    else:
                        self._get_csv_from_s3(config)
                    config.config.set_last_run_time(utcnow, None, True)
                else:
                    self.logger.info('Billing thread %s is disabled',
                                     config.section_name)

            #pylint: disable=bare-except
            except:
                self.logger.error('%s failed: %s', config.section_name,
                                  sys.exc_info()[1])
                traceback.print_exc()

    def _get_csv_from_local(self, config):
        """
        Opens a CSV file on the local machine
        Arguments:
        config - the AwsBillingDetailThreadConfiguration object
        """

        self.logger.info('Getting AWS billing details from local file system %s',
                         config.section_name)
        with open(config.prefix, 'r') as csvfd:
            csv_file = utils.CsvFile(csvfd, config.header_row_index)
            self.parse_csv(config, csv_file, 'local')

    #pylint: disable=too-many-locals
    #pylint: disable=too-many-branches
    #pylint: disable=too-many-statements
    def _get_csv_from_s3(self, config):
        """
        Opens a CSV file that matches the prefix in the S3 bucket.
        Arguments:
        config - the AwsBillingDetailThreadConfiguration object
        """

        self.logger.info('Getting AWS billing details from S3 for %s',
                         config.section_name)
        utcnow = (datetime.datetime.utcnow()
                  .replace(microsecond=0, tzinfo=dateutil.tz.tzutc()))

        s3cli = self.account.get_session(
            config.region, self.config.role_arn,
            self.config.role_external_id).client('s3')
        acct_id = self.account.get_account_id(self.config.role_arn)
        curr_month = utcnow.strftime('%Y-%m')
        prefix = (config.prefix
                  .replace('${account_id}', acct_id)
                  .replace('${date}', curr_month))

        # find the item in the s3 bucket
        response = s3cli.list_objects(Bucket=config.bucket, Prefix=prefix)
        if (not response or 'Contents' not in response or
                not response['Contents']):
            self.logger.warning('Billing details file [%s] not found in %s\n%s',
                                prefix, config.bucket, str(response))
            return

        # open the item in S3
        key = None
        zipped = False
        for s3file in response['Contents']:
            if s3file['Key'][-8:] == '.csv.zip':
                key = s3file['Key']
                zipped = True
                break

            if s3file['Key'][-4:] == '.csv':
                key = s3file['Key']
                zipped = False

        if not key:
            self.logger.warning('Unable to find billing file [%s] in %s',
                                prefix, config.bucket)
            return

        try:
            response = s3cli.get_object(Bucket=config.bucket, Key=key)
            csv_contents = io.BytesIO(response['Body'].read())

            filename = None
            if zipped:
                self.logger.info('Unzipping %s ...', key)
                with zipfile.ZipFile(csv_contents, 'r') as zipref:
                    key = key[0:-4] # remove .zip
                    filename = config.tmp_dir + key
                    zipref.extractall(config.tmp_dir)

                if utils.CANCEL_WORKERS_EVENT.is_set():
                    return
                with open(filename, 'r') as csvfd:
                    csv_file = utils.CsvFile(csvfd, config.header_row_index)
                    self.parse_csv(config, csv_file, curr_month)

            else:
                csv_file = utils.CsvFile(csv_contents, config.header_row_index)
                self.parse_csv(config, csv_file, curr_month)

        finally:
            if filename and os.path.exists(filename):
                self.logger.info('Removing %s ...', filename)
                os.remove(filename)

    def parse_csv(self, config, csvreader, curr_month):
        """
        Parse the CSV contents and generate metrics.
        Arguments:
        config - the AwsBillingDetailThreadConfiguration object
        csvreader - utils.CsvFile object
        curr_month - Y-M
        """

        rows = 0
        record_id = None
        current_record_id = None
        if config.record_id_column:
            record_id = config.get_last_record_id(curr_month)
            self.logger.info('Skipping records until after record ID %s', record_id)

        # loop over all lines in the csv file after the header and
        # transmit the cost metric for each one
        #pylint: disable=bare-except
        for row in csvreader:
            if utils.CANCEL_WORKERS_EVENT.is_set():
                break

            try:
                if config.record_id_column and row[config.record_id_column]:
                    current_record_id = row[config.record_id_column]
                    if record_id and current_record_id != record_id:
                        continue
                    elif record_id and current_record_id == record_id:
                        record_id = None
                        continue
                    else:
                        record_id = None

                self._process_csv_row(row, config)

            except:
                self.logger.warning('Unable to process record (%s):\n\t%s',
                                    sys.exc_info()[1], str(row))
                traceback.print_exc()

            rows = rows + 1
            if config.maximum_number_of_rows:
                if rows >= config.maximum_number_of_rows:
                    self.logger.debug('Stopping after %d rows', rows)
                    break

            if config.sleep_after_rows and rows % config.sleep_after_rows == 0:
                self.logger.debug('Sleeping %0.2f', config.sleep_ms)
                time.sleep(config.sleep_ms)

        if current_record_id:
            self.logger.info('Recording last record id of %s for %s',
                             current_record_id, curr_month)
            config.set_last_record_id(curr_month, current_record_id)

    def _process_csv_row(self, row, config):

        # point tags
        point_tags = {}
        for header, point_tag_key in config.dimensions.iteritems():
            if row[header]:
                point_tags[point_tag_key] = row[header]

        # point tags from ec2 instance
        #pylint: disable=too-many-nested-blocks
        if config.instance_id_columns:
            found_instance = False
            for header in config.instance_id_columns:
                instance_id = row[header]

                # arn:aws:ec2:us-east-1:011750033084:instance/i-33ac36e5"
                if instance_id and instance_id[0:12] == 'arn:aws:ec2:':
                    parts = instance_id.split(':')
                    instance_id = parts[5].split('/')[1]
                    point_tags['region'] = parts[3]

                if not instance_id or instance_id[0:2] != 'i-':
                    continue
                for region in self.account.regions:
                    for sub_account in self.account.get_sub_accounts():
                        instances = sub_account.get_instances(region)
                        if instance_id in instances:
                            instance_tags = instances[instance_id]
                            for key, value in instance_tags.iteritems():
                                point_tags[key] = value
                            found_instance = True
                            break
                    if found_instance:
                        break

                if found_instance:
                    break

        # source names
        source, source_name = AwsBaseMetricsCommand.get_source(
            config.source_names, point_tags)
        if source_name in point_tags:
            del point_tags[source_name]

        # timestamp
        tstamp = None
        tstamp_col_values = []
        for header, date_fmt in config.dates.iteritems():
            if row[header]:
                tstamp_col_values.append(row[header])
                tstamp = utils.unix_time_seconds(
                    datetime.datetime.strptime(row[header], date_fmt))

        if not tstamp:
            self.logger.warning('Unable to find valid date in columns (%s) '
                                '|%s|.  Record is:\n\t%s',
                                ', '.join(config.dates.keys()),
                                ', '.join(tstamp_col_values),
                                str(row))
            return

        # calculate duration
        if config.duration and len(config.duration) == 2:
            start = config.duration[0].split('|')
            start_dt = datetime.datetime.strptime(row[start[0]],
                                                  start[1])
            start_tstamp = utils.unix_time_seconds(start_dt)

            end = config.duration[1].split('|')
            end_dt = datetime.datetime.strptime(row[end[0]], end[1])
            end_tstamp = utils.unix_time_seconds(end_dt)

            duration = end_tstamp - start_tstamp
        else:
            duration = 0

        # metric and value
        for header, metric_name in config.metrics.iteritems():
            if config.namespace:
                metric = config.namespace + '.' + metric_name
            else:
                metric = metric_name

            value = row[header]
            if not value:
                value = 0.0

            # send the metric to the proxy
            self.proxy.transmit_metric(metric, value, long(tstamp),
                                       source, point_tags)
            if duration:
                self.proxy.transmit_metric(metric + '.duration',
                                           duration, long(tstamp),
                                           source, point_tags)

